from array import array
import logging
import os
import struct
import sys
import Image # from Python Imaging Library

from agos_util import *

DUMP_RAW = False

class VgaFile(object):
    def __init__(self):
        pass
    
    def seek(self, offset, whence=os.SEEK_SET):
        self.f_handle.seek(offset, whence)

    def read(self, size=None):
        return self.f_handle.read(size)

    def tell(self):
        return self.f_handle.tell()

    def close(self):
        self.f_handle.close()

class VgaFileOldBundle(VgaFile):
    def __init__(self, input_file):
        self.f_handle = open(input_file, 'rb')
        self.block_start = 0
        self.f_handle.seek(0, 2)
        self.block_end = self.f_handle.tell()
        self.f_handle.seek(0, 0)

def openVgaFilesForZone(zoneNum):
    use_old_bundle = True
    if use_old_bundle:
        fname = "%s.vga" % (str(zoneNum).zfill(3) + "1")
        vga1_file = VgaFileOldBundle(fname)
        fname = "%s.vga" % (str(zoneNum).zfill(3) + "2")
        vga2_file = VgaFileOldBundle(fname)
    else:
        #VgaFileNewBundle
        raise NotImplementedException("No support for new bundles")
    return vga1_file, vga2_file

def palLoad(vga1_file, pal_num, dump_full_palette):
    # palNum is always 0 when called from dumpVgaBitmaps - leads to some
    #  images having the wrong palette. I think the palette is chosen within VGA
    #  scripts.
    vga_pal = array('B')
    out_pal = array('B')

    if dump_full_palette: # doesn't work - reads values that overflow, possibly not palette.
        num = 256
    else:
        num = 32 # number of colours in the palette
    palSize = num * 3 # total size of the palette

    if dump_full_palette:
        vga1_file.seek(6)
    else:
        vga1_file.seek(6 + pal_num * palSize)
    vga_pal.fromfile(vga1_file.f_handle, palSize)

    # convert from VGA palette
    for i in xrange(len(vga_pal)):
        out_pal.append(vga_pal[i] << 2)

    return out_pal

def depackColumn(input_file, height, reps):
    """
    Reps can be carried over between columns. e.g. reps is 5, column A only has 3
    values, so the remaining 2 values are read from column B.

    If reps is < 0, it represents the inverse number of non-repeated values that
    follow. e.g. if reps is -5, read in & output 5 bytes.
    """
    col_data = array('B')
    if reps == -0x80:
        reps = readSignedByte(input_file)

    while True:
        if reps >= 0:
            colour = readByte(input_file)
            while True:
                col_data.append(colour)
                height -= 1
                if height == 0:
                    reps -= 1
                    if reps < 0:
                        reps = -0x80
                    else:
                        input_file.seek(-1, os.SEEK_CUR) # hmm not sure about this
                    return col_data, reps # might need to return src, a, and "depack_dest + y_skip"
                reps -= 1
                if reps < 0:
                    break
        else:
            while True:
                val = readByte(input_file)
                col_data.append(val)

                height -= 1
                if height == 0:
                    reps += 1
                    if reps == 0:
                        reps = -0x80
                    return col_data, reps # might need to return src, a, and "depack_dest + y_skip"
                
                reps += 1
                if reps == 0:
                    break
        reps = readSignedByte(input_file)


def dumpBitmap(out_file_name, input_file, w, h, flags, palette, base):
    imageBuffer = array('B')

    reps = -0x80
    # TODO?
#    if (w > _screenWidth):
#        raise NotSupportedError("Compression not supported.")
#    elif (h > _screenHeight):
#        raise NotSupportedError("Compression not supported.")
#    elif
    if (w == 320) and (h == 134 or h == 200):
        logging.debug("Decoding background with 8-bytes-in-5 algorithm")
        # This compression technique seems to take advantage of the fact that
        #  each Simon image only uses 32 colours in the palette.
        #  It manages to cram 8 bits into 5.
        for _ in xrange(h):
            count = w / 8

            while True:
                bits = readUint32(input_file) # I think this is big endian... ScummVM has manual reading in?
                for shift_diff in xrange(1, 7):
                    val = ((bits >> (32 - (5 * shift_diff))) & 31)
                    imageBuffer.append(val)

                bits = (bits << 8) | readByte(input_file)

                val = ((bits >> (40 - 35)) & 31)
                imageBuffer.append(val)
                val = (bits & 31)
                imageBuffer.append(val)

                count -= 1
                if count == 0:
                    break
    elif flags & 0x80:
        logging.debug("Decoding 16-colour 4-bit with column RLE algorithm")
        # Need to initialise the whole image buffer, since we're not appending
        # values per row.
        imageBuffer.fromstring('\x00' * w * h)
        #print len(imageBuffer)
        for i in xrange(0, w, 2):
            col_data, reps = depackColumn(input_file, h, reps)
            for j in xrange(h):
                col_val = col_data[j]
                val = (col_val >> 4) | base
                imageBuffer[j * w + i] = val
                val = (col_val & 0xF) | base
                imageBuffer[j * w + i + 1] = val
    else:
        logging.debug("Decoding 16-colour 4-bit uncompressed algorithm")
        for _ in xrange(h):
            for _ in xrange(w / 2):
                col = readByte(input_file)
                val = col >> 4
                imageBuffer.append(val)
                val = col & 0xF
                imageBuffer.append(val)

    saveToPng(out_file_name, palette, imageBuffer, w, h)
    if DUMP_RAW:
        _saveRaw(out_file_name, imageBuffer, w, h)

def dumpVgaBitmaps(zoneNum, dump_full_palette=False):
    try:
        vga1_file, vga2_file = openVgaFilesForZone(zoneNum)
    except Exception, e:
        logging.debug("Could not open any VGA files for zone: %s. Reason: %s" % (zoneNum, e))
        return

    # Size of the image data in the second VGA file
    imageBlockSize = vga2_file.block_end

    pal_num = 0
    palette_array = palLoad(vga1_file, pal_num, dump_full_palette)

    i = 1
    currOffs = 0
    while vga2_file.tell() < vga2_file.block_end: # was an endless loop
        # VGA header is 8 bytes, format:
        #  - offset = 32 BE
        #  - flags = 8
        #  - height = 8
        #  - width = 16 BE
        vga2_file.seek(i * 8)
        data = vga2_file.read(8)
        offs, flags, height, width = struct.unpack('>I2BH', data)

        logging.debug("Zone %d: Image %d, Offs=%d Width=%d, Height=%d, Flags=%s" %
            (zoneNum, i, offs, width, height, myhex(flags)))
        if (offs < currOffs or offs >= imageBlockSize or width == 0 or height == 0):
            return

        currOffs = offs
        outName = os.path.join("dumps", "Res%d_Image%d.png" % (zoneNum, i))

        # Seek to the location of the bitmap data for this image.
        vga2_file.seek(offs)
        dumpBitmap(outName, vga2_file, width, height, flags, palette_array, 0)

        i += 1
    pass

def saveToPng(image_path, palette, bitmap, width, height):
    # Create our new image!
    im = Image.new('P', (width, height) )
    im.putpalette(palette)
    im.putdata(bitmap)
    if os.path.splitext(image_path)[1].lower() != '.png':
        image_path += '.png'
    im.save(image_path, 'png') # always saves to PNG files.
    logging.debug("Image saved to: %s" % image_path)

def _convertTo4Bit(sourceData, width, height):
    imageBuffer = array('B')
    for row_i in xrange(height):
        for col_i in xrange(0, width, 2):
            col_1 = sourceData[row_i * width + col_i]
            col_2 = sourceData[row_i * width + col_i + 1]
            val = ((col_1 & 0xF) << 4) | (col_2 & 0xF)
            imageBuffer.append(val)
    return imageBuffer

def _convertToColumns(sourceData, width, height):
    imageBuffer = array('B')
    for col_i in xrange(width):
        for row_i in xrange(height):
            val_index = row_i * width + col_i
            val = sourceData[val_index]
            imageBuffer.append(val)
    return imageBuffer

def _saveRaw(image_path, bitmap, width, height):
    # Mostly for debugging
    out_file = file(image_path + ".raw", 'wb')
    bitmap.tofile(out_file)
    out_file.close()
    bitmap = _convertTo4Bit(bitmap, width, height)
    bitmap = _convertToColumns(bitmap, width / 2, height)
    out_file = file(image_path + ".col4.raw", 'wb')
    bitmap.tofile(out_file)
    out_file.close()

def _test(args):
    # TODO: option to output the entire palette
    logging.basicConfig(format="", level=logging.DEBUG)
    if len(args) >= 2 or len(args) == 0:
        #logging.error("Please only pass one zone ID as the argument, and 'Y' if you want to include the full palette in the output file.")
        logging.error("Please only pass one zone ID as the argument.")
        sys.exit(-1)
#    elif len(args) == 2:
#        if args[1].lower() == 'y':
#            dump_full_palette = True
#        else:
#            dump_full_palette = False
    elif len(args) == 1:
        dump_full_palette = False
    dumpVgaBitmaps(int(args[0]), dump_full_palette)
    sys.exit(0)

if __name__ == "__main__": _test(sys.argv[1:])
