# To change this template, choose Tools | Templates
# and open the template in the editor.
from array import array
import logging
import sys
import Image # from Python Imaging Library

from agos_util import EncoderError, NotSupportedError, NotImplementedError

COMP_4_BIT = 1
COMP_4_BIT_COL_RLE = 2
COMP_8_IN_5_BKGD = 3

def _writeValBuffer(imageBuffer, val_buffer):
    imageBuffer.append(-len(val_buffer) & 0xFF)
    for v in val_buffer:
        imageBuffer.append(v & 0xFF)
    del val_buffer[:]

def packColumns(sourceData, width, height):
    reps = 0
    val_buffer = []
    in_run = False

    imageBuffer = array('B')

    for i in xrange(len(sourceData)):
        val = sourceData[i]
        # Reached EOF?
        if i == len(sourceData) - 1:
            next_val = None
        # Get next value
        else:
            next_val = sourceData[i + 1]

        # If the next value is not the same as the current value,
        #  - in run: write out what we've got so far, reset counter.
        #  - not in run: add value to buffer of non-repeating values.
        if next_val != val:
            if in_run:
                imageBuffer.append(reps & 0xFF)
                imageBuffer.append(val & 0xFF)
                reps = 0
                in_run = False
            else:
                val_buffer.append(val)
                # Write buffer out if we've reached the end of the file,
                #  or the buffer is too big
                if next_val is None or len(val_buffer) == 127:
                    _writeValBuffer(imageBuffer, val_buffer)
        # If next and current values are the same,
        # - in run: increment run counter
        # - the buffer of non-repeating values.
        else:
            if in_run:
                # If already in run, continue the run.
                reps += 1
            else:
                # If not in a run, start a new run. If we have any non-reapeating
                #  values buffered, write them out.
                if len(val_buffer) > 0:
                    _writeValBuffer(imageBuffer, val_buffer)
                in_run = True
                reps = 1
    return imageBuffer

def convertTo4Bit(sourceData, width, height):
    imageBuffer = array('B')
    for row_i in xrange(height):
        for col_i in xrange(0, width, 2):
            col_1 = sourceData[row_i * width + col_i]
            col_2 = sourceData[row_i * width + col_i + 1]
            val = ((col_1 & 0xF) << 4) | (col_2 & 0xF)
            imageBuffer.append(val)
    return imageBuffer

def convertToColumns(sourceData, width, height):
    imageBuffer = array('B')
    for col_i in xrange(width):
        for row_i in xrange(height):
            val_index = row_i * width + col_i
            val = sourceData[val_index]
            imageBuffer.append(val)
    return imageBuffer

def encodeBitmap(source, compressionMethod, base=0):
    sourceData = source.getdata()
    width, height = source.size
    imageBuffer = None

    if compressionMethod == COMP_4_BIT:
        logging.debug("Encoding 16-colour 4-bit uncompressed algorithm")
        imageBuffer = convertTo4Bit(sourceData, width, height)

    elif compressionMethod == COMP_4_BIT_COL_RLE:
        logging.debug("Encoding 16-colour 4-bit with column RLE algorithm")
        # First compress into 4-bit, then convert to columns, then compress with RLE.
        # (do it this way because "reps" can carry across mutliple columns)
        image_data_4bit = convertTo4Bit(sourceData, width, height)
        columnised_data = convertToColumns(image_data_4bit, width / 2, height)
        result = packColumns(columnised_data, width / 2, height)
        imageBuffer = result

    elif compressionMethod == COMP_8_IN_5_BKGD:
        raise NotImplementedError("Unfotunately this compression has not been implemented yet.")

    else:
        raise NotSupportedError("Unsupported compression format: %s" % str(compressionMethod))

    return imageBuffer

def saveToRaw(out_file_name, bitmap):
    out_file = file(out_file_name, 'wb')
    bitmap.tofile(out_file)
    out_file.close()

def main(args):
    logging.basicConfig(format="", level=logging.DEBUG)
    if len(args) != 2:
        print "Please pass the input file name and the encoding method (1 = 4 bit, 2 = 4 bit with column RLE, 3 = 5 bit)"
        return 1

    image_path = args[0]
    compressionMethod = int(args[1])
    source = Image.open(image_path)
    # Validation
    min_value, max_value = source.getextrema()
    if source.palette == None:
        raise EncoderError("Only images with a palette are supported (i.e. 8-bit or less).")
    if (compressionMethod in (COMP_4_BIT, COMP_4_BIT_COL_RLE) and
        max_value > 16):
        raise EncoderError("4-bit compression modes cannot use palette values > 16.")
    elif (compressionMethod == COMP_8_IN_5_BKGD and
            max_value > 32):
        raise EncoderError("5-bit compression modes cannot use palette values > 32.")

    resultBuffer = encodeBitmap(source, compressionMethod)
    out_name = image_path + ".out"
    saveToRaw(out_name, resultBuffer)
    print "Done, saved as %s." % out_name

if __name__ == "__main__": main(sys.argv[1:])

